//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "SignalCfg.h"
#include "SignalSynthCfg.h"
#include "ArmScope.h"
#include "armscope_cmd.h"
#include "Utils.h"
#include <json/json.h>
#include <fstream>
#include <SysUtils.hpp>
#include <algorithm>

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TfrmSignalConfig *frmSignalConfig = NULL;
//---------------------------------------------------------------------------

__fastcall TfrmSignalConfig::TfrmSignalConfig(TComponent* Owner): TForm(Owner),
		running(false),
		mouseDown(false)
{
	reqFreq = 20000.0f;
	edDacFreq->Text = FloatToStrF(reqFreq, ffFixed, 6, 2);
	SetNearestRealFreq(reqFreq);

	//image->Canvas->Brush->Color = clWhite;
	image->Canvas->FloodFill(0, 0, clWhite, fsBorder);

	UpdateSamplesInfo();
}

void TfrmSignalConfig::SetNearestRealFreq(float freq)
{
	// base frequency = 72 MHz
	STM32_prescaler = 0;
	STM32_reload = 72000000.0f / freq;
	realFreq = 72000000.0f / STM32_reload;
	edDacRealFreq->Text = FloatToStrF(realFreq, ffFixed, 6, 2);
}

void __fastcall TfrmSignalConfig::btnStartClick(TObject *Sender)
{
	UpdateFreq();
	
	struct ThreadComm::Cmd cmd;
	memset(&cmd, 0, sizeof(cmd));
	cmd.buffer[0] = CMD_DAC;
	cmd.buffer[1] = CMD_DAC_START;
	if (armScope.SendCmd(cmd) == 0)
		running = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmSignalConfig::btnStopClick(TObject *Sender)
{
	struct ThreadComm::Cmd cmd;
	memset(&cmd, 0, sizeof(cmd));
	cmd.buffer[0] = CMD_DAC;
	cmd.buffer[1] = CMD_DAC_STOP;
	armScope.SendCmd(cmd);
	running = false;
}
//---------------------------------------------------------------------------


void __fastcall TfrmSignalConfig::edDacFreqChange(TObject *Sender)
{
	AnsiString val = edDacFreq->Text;
	float fVal;
	try
	{
		fVal = StrToFloat(val);
	}
	catch(...)
	{
		return;
	}
	if (fVal > 1000000.0f || fVal < 0.01f)
	{
		ShowMessage("Frequency out of allowed range!");
		return;
	}
	reqFreq = fVal;
	SetNearestRealFreq(fVal);
}
//---------------------------------------------------------------------------

void TfrmSignalConfig::UpdateFreq(void)
{
	struct ThreadComm::Cmd cmd;
	memset(&cmd, 0, sizeof(cmd));
	cmd.buffer[0] = CMD_DAC;
	cmd.buffer[1] = CMD_DAC_CLOCK_CONFIG;
	memcpy(&cmd.buffer[2], &STM32_reload, sizeof(STM32_reload));
	memcpy(&cmd.buffer[2] + sizeof(STM32_reload), &STM32_prescaler, sizeof(STM32_prescaler));
	armScope.SendCmd(cmd);

	if (running)
	{
		memset(&cmd, 0, sizeof(cmd));
		cmd.buffer[0] = CMD_DAC;
		cmd.buffer[1] = CMD_DAC_STOP;
		armScope.SendCmd(cmd);

		memset(&cmd, 0, sizeof(cmd));
		cmd.buffer[0] = CMD_DAC;
		cmd.buffer[1] = CMD_DAC_START;
		armScope.SendCmd(cmd);
	}
}

void __fastcall TfrmSignalConfig::btnUpdateFreqClick(TObject *Sender)
{
	UpdateFreq();	
}
//---------------------------------------------------------------------------

void __fastcall TfrmSignalConfig::btnFreq1ppClick(TObject *Sender)
{
	reqFreq *= 1.01f;
	SetNearestRealFreq(reqFreq);
	edDacFreq->OnChange = NULL;
	edDacFreq->Text = FloatToStrF(reqFreq, ffFixed, 6, 2);
	edDacFreq->OnChange = edDacFreqChange;
}
//---------------------------------------------------------------------------

void __fastcall TfrmSignalConfig::btnFreq1pmClick(TObject *Sender)
{
	reqFreq /= 1.01f;
	SetNearestRealFreq(reqFreq);
	edDacFreq->OnChange = NULL;
	edDacFreq->Text = FloatToStrF(reqFreq, ffFixed, 6, 2);
	edDacFreq->OnChange = edDacFreqChange;
}
//---------------------------------------------------------------------------

void __fastcall TfrmSignalConfig::btnFreq10ppClick(TObject *Sender)
{
	reqFreq *= 1.1f;
	SetNearestRealFreq(reqFreq);
	edDacFreq->OnChange = NULL;
	edDacFreq->Text = FloatToStrF(reqFreq, ffFixed, 6, 2);
	edDacFreq->OnChange = edDacFreqChange;
}
//---------------------------------------------------------------------------

void __fastcall TfrmSignalConfig::btnFreq10pmClick(TObject *Sender)
{
	reqFreq /= 1.1f;
	SetNearestRealFreq(reqFreq);
	edDacFreq->OnChange = NULL;
	edDacFreq->Text = FloatToStrF(reqFreq, ffFixed, 6, 2);
	edDacFreq->OnChange = edDacFreqChange;
}
//---------------------------------------------------------------------------


void __fastcall TfrmSignalConfig::imageMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
	mouseDown = true;
	startX = X;
	startY = Y;
	image->Canvas->Brush->Color = clWhite;
	image->Canvas->FillRect(Rect(X, 0, X+1, image->Height));
	image->Canvas->Brush->Color = clBlack;
	image->Canvas->Pixels[X][Y] = clBlack;
	AddSamplesPoint(X, Y);		
}
//---------------------------------------------------------------------------

void __fastcall TfrmSignalConfig::imageMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
	mouseDown = false;	
}
//---------------------------------------------------------------------------

void __fastcall TfrmSignalConfig::imageMouseMove(TObject *Sender,
      TShiftState Shift, int X, int Y)
{
	if (mouseDown)
	{
		// erase previous content in this time span
		image->Canvas->Brush->Color = clWhite;
		image->Canvas->FillRect(Rect(std::min(X, startX), 0, std::max(X, startX), image->Height));
		image->Canvas->Brush->Color = clBlack;

		if (X == startX)
		{
			image->Canvas->Pixels[X][Y] = clBlack;
			AddSamplesPoint(X, Y);
		}
		else
		{
			image->Canvas->MoveTo(startX, startY);
			image->Canvas->LineTo(X, Y);
			AddSamplesArea(startX, startY, X, Y);
		}
		startX = X;
		startY = Y;
	}
	AnsiString statusText;
	statusText.sprintf("X: %4.4f ms, Y: %1.2f V",
		(float)X/realFreq, (float)(255-Y)/256 * 3.3);
	StatusBar->SimpleText = statusText;
}
//---------------------------------------------------------------------------

void __fastcall TfrmSignalConfig::btnSetWaveClick(TObject *Sender)
{
	UpdateShape();
}
//---------------------------------------------------------------------------

void TfrmSignalConfig::UpdateShape(void)
{
	struct ThreadComm::Cmd cmd;
	memset(&cmd, 0, sizeof(cmd));
	cmd.buffer[0] = CMD_DAC;

	if (running)
	{
		cmd.buffer[1] = CMD_DAC_STOP;
		armScope.SendCmd(cmd);
	}

	cmd.buffer[1] = CMD_DAC_CLEAR_WAVE;
	armScope.SendCmd(cmd);

	unsigned int samplesLeft = samples.size();
	cmd.buffer[1] = CMD_DAC_SET_WAVE;
	while (samplesLeft > 0)
	{
		cmd.buffer[2] = std::min(samplesLeft, sizeof(cmd.buffer)-3);
		memcpy(&cmd.buffer[3], &samples[samples.size()-samplesLeft], cmd.buffer[2]);
		samplesLeft -= cmd.buffer[2];
		if (armScope.SendCmd(cmd))
		{
        	break;
		}
	}

	if (running)
	{
		cmd.buffer[1] = CMD_DAC_START;
		armScope.SendCmd(cmd);
	}
}

void __fastcall TfrmSignalConfig::btnClearPlotClick(TObject *Sender)
{
	image->Canvas->Brush->Color = clWhite;
	image->Canvas->FillRect(image->Canvas->ClipRect);
	image->Canvas->Brush->Color = clBlack;
	samples.clear();
	UpdateSamplesInfo();
}
//---------------------------------------------------------------------------

void TfrmSignalConfig::UpdateSamplesInfo(void)
{
	AnsiString text;
	text.sprintf("Samples count: %d", samples.size());
	lblShapeSamples->Caption = text;
}

void TfrmSignalConfig::AddSamplesPoint(int x, int y)
{
	unsigned char lastVal;

	if (samples.size() == 0)
	{
		lastVal = PlotYToVal(y);
	}
	else
	{
		lastVal = samples[samples.size()-1];
	}
	int yPos = ValToPlotY(lastVal);

	// fill the gap
	TCanvas *c = image->Canvas;
	for (int i=samples.size(); i<x; i++)
	{
		samples.push_back(lastVal);
		c->Pixels[i][yPos] = clBlack;
	}

	// add or update value
	if (x >= (int)samples.size())
	{
		samples.push_back(PlotYToVal(y));
	}
	else
	{
		samples[x] = PlotYToVal(y);
	}
	UpdateSamplesInfo();
}

void TfrmSignalConfig::AddSamplesArea(int x1, int y1, int x2, int y2)
{
	int Xmin, Xmax, Ymin, Ymax;
	Xmin = std::min(x1, x2);
	Xmax = std::max(x1, x2);
	Ymin = std::min(y1, y2);
	Ymax = std::max(y1, y2);

	if (Xmin == Xmax)
		return;

	unsigned char lastVal;
	if (samples.size() == 0)
	{
		lastVal = PlotYToVal(y1);
	}
	else
	{
		lastVal = samples[samples.size()-1];
	}
	int yPos = ValToPlotY(lastVal);

	TCanvas *c = image->Canvas;

	// fill the gap
	for (int i=samples.size(); i<Xmin; i++)
	{
		samples.push_back(lastVal);
		c->Pixels[i][yPos] = clBlack;
	}

	for (int x=Xmin; x<Xmax; x++)
	{
		for (int y=Ymin; y<=Ymax; y++)
		{
			TColor color = c->Pixels[x][y];
			if (color == clBlack)
			{
				// add or update value
				if (x >= (int)samples.size())
				{
					samples.push_back(PlotYToVal(y));
				}
				else
				{
					samples[x] = PlotYToVal(y);
				}
				break;
			}
		}
	}

	UpdateSamplesInfo();
}


void __fastcall TfrmSignalConfig::miSaveJsonClick(TObject *Sender)
{
	saveDialog->DefaultExt = "json";
	saveDialog->Filter = "JSON file (*.json)|*.json";
	saveDialog->FilterIndex = 1;
	if (saveDialog->Execute())
	{
		Json::Value root;
		Json::StyledWriter writer;

		Json::Value &wave = root["Wave"];
		wave["DAC feeding frequency"] = reqFreq;
		Json::Value &jsamples = wave["Samples"];
		for (unsigned int i=0; i<samples.size(); i++)
		{
			jsamples[i] = samples[i];
		}

		std::string output = writer.write( root );
		std::ofstream ofs(saveDialog->FileName.c_str());
		if (!ofs.is_open())
		{
			ShowMessage("Could not write file");
			return;
		}

		ofs << output;
		ofs.close();
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmSignalConfig::miOpenJsonClick(TObject *Sender)
{
	openDialog->DefaultExt = "json";
	openDialog->Filter = "JSON file (*.json)|*.json";
	openDialog->FilterIndex = 1;
	if (openDialog->Execute())
	{
		AnsiString filename = openDialog->FileName;
		std::ifstream ifs(filename.c_str());
		if (!ifs.is_open())
		{
			ShowMessage("Could not open file");
			return;
		}
		std::string strJson((std::istreambuf_iterator<char>(ifs)), std::istreambuf_iterator<char>());
		ifs.close();

		Json::Value root;   // will contains the root value after parsing.
		Json::Reader reader;
		bool parsingSuccessful = reader.parse( strJson, root );
		if ( !parsingSuccessful )
		{
			ShowMessage("Could not parse JSON file");
			return;
		}

		Json::Value &Wave = root["Wave"];
		if (Wave.type() != Json::objectValue)
		{
			ShowMessage("Incorrect file structure");
			return;
		}

		Json::Value &dacFeedingFreq = Wave["DAC feeding frequency"];
		if (dacFeedingFreq.type() != Json::realValue)
		{
			ShowMessage("Incorrect file structure");
			return;
		}
		reqFreq = dacFeedingFreq.asDouble();
		SetNearestRealFreq(reqFreq);
		edDacFreq->OnChange = NULL;
		edDacFreq->Text = FloatToStrF(reqFreq, ffFixed, 6, 2);
		edDacFreq->OnChange = edDacFreqChange;
      	UpdateFreq();

		Json::Value &jsamples = Wave["Samples"];
		if (jsamples.type() != Json::arrayValue)
		{
			ShowMessage("Incorrect file structure");
			return;
		}

		samples.clear();
		for (unsigned int i=0; i<jsamples.size(); i++)
		{
        	samples.push_back(jsamples[i].asDouble());
		}

		image->Canvas->Brush->Color = clWhite;
		image->Canvas->FillRect(image->Canvas->ClipRect);
		image->Canvas->Brush->Color = clBlack;
		for (unsigned int i=0; i<samples.size(); i++)
		{
        	image->Canvas->Pixels[i][ValToPlotY(samples[i])] = clBlack;
		}

		UpdateSamplesInfo();
		UpdateShape();	

	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmSignalConfig::pictboxVertScalePaint(TObject *Sender)
{
	int width = pictboxVertScale->Width;
	TCanvas* canvas = pictboxVertScale->Canvas;
	int offset = 2;
	for (int i=0; i<15; i++)
	{
		canvas->MoveTo(0, 15+i*16 + offset);
		canvas->LineTo(width, 15+i*16 + offset);
		if (i == 7) {
			canvas->MoveTo(0, 14+i*16 + offset);
			canvas->LineTo(width, 14+i*16 + offset);
			canvas->MoveTo(0, 16+i*16 + offset);
			canvas->LineTo(width, 16+i*16 + offset);
		}
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmSignalConfig::btnStandardShapesClick(TObject *Sender)
{
	if (frmSignalSynthConfig == NULL) {
    	frmSignalSynthConfig = new TfrmSignalSynthConfig(this);
	}
	frmSignalSynthConfig->fBaseSamplingFreq = realFreq;
	frmSignalSynthConfig->iMaxSamples = image->Width;
	frmSignalSynthConfig->ShowModal();
	if (frmSignalSynthConfig->Confirmed())
	{
		int count = 0;
		short *data = frmSignalSynthConfig->Generate(3.3f/256.0f, count);
		if (data)
		{
			samples.clear();
			for (int i=0; i<count; i++)
			{
				if (data[i] < 0)
				{
					samples.push_back(0);
				}
				else if (data[i] > 255)
				{
					samples.push_back(255);
				}
				else
				{
					samples.push_back(data[i]);
				}
			}

			image->Canvas->Brush->Color = clWhite;
			image->Canvas->FillRect(image->Canvas->ClipRect);
			image->Canvas->Brush->Color = clBlack;
			for (unsigned int i=0; i<samples.size(); i++)
			{
				image->Canvas->Pixels[i][ValToPlotY(samples[i])] = clBlack;
			}

			UpdateSamplesInfo();
			UpdateShape();
		}
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmSignalConfig::imageMouseLeave(TObject *Sender)
{
	StatusBar->SimpleText = "";	
}
//---------------------------------------------------------------------------

